#include "typedef.h"

#include <stdio.h>

#include "util.h"
#include "object.h"
#include "class.h"

Cclass	*Cclass::g_hash[HASHSIZE] = {0};

/* static */ u4T
Cclass::hash(const u2T *nameP)
{
	int		ret  = 0;
	u2T		c;
	const	u2T *P;

	for (P = nameP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	return(ret & (HASHSIZE-1));
}

Cclass::Cclass(const u2T *nameP, int dimensions)
{
	m_nameP            = unicodedup(nameP);

	m_parentP          = 0;
	m_hashP            = 0;
	m_major_version    = 0;
	m_minor_version    = 0;
	m_access_flags     = 0;
	m_extendsP         = 0;
	m_implements_cnt   = 0;
	m_implementsPP     = 0;
	m_innerclasses_cnt = 0;
	m_innerclassesP    = 0;
	m_arrayofP         = 0;
	m_lineno           = -1;
	m_dimensions       = dimensions;
	m_processed        = 0;
}

void
Cclass::set_parent(Centity *parentP)
{
	m_parentP = parentP;
	add_created();
}

Cclass *
Cclass::locate(u2T *nameP, int dimensions, int datatype)
{
	unsigned int	index;
	Cclass			**classPP, *classP;
	u2T				c, *P;

	P = nameP;
	c = *P;
	if (c == 'L') {
		for (; c = *P; ++P) {
			if (c == ';') {
				++nameP;
				break;
		}	}
		*P = 0;
	}	

	index = hash(nameP);
	for (classPP = g_hash + index; ; classPP = &(classP->m_hashP)) {

		if (!(classP = *classPP)) {
			classP = new Cclass(nameP, dimensions);
			if (!classP) {
				outofmemory();
			}
			*classPP = classP;
			if (dimensions) {
				classP->m_arrayofP = locate(nameP, dimensions-1, datatype);
			}
			if (datatype) {
				classP->m_parentP = (Centity *) -1;
			}
			break;
		}

		if (unicodecmp(classP->m_nameP, nameP)) {
			continue;
		}
		if (classP->m_dimensions != dimensions) {
			continue;
		}
		break;
	}
	*P = c;
	return(classP);
}

Cclass *
Cclass::locate(u2T *nameP)
{
	int		dimensions = 0;
	Cclass	*classP;
	u2T		*P;
	int		datatype;

	datatype = 0;
	dimensions = 0;
	for (P = nameP; *P == '['; ++P) {
		++dimensions;
	}
	if (dimensions && *P != 'L') {
		P = get_datatype(*P);
		datatype = 1;
	}
	classP = locate(P, dimensions, datatype);
	return(classP);
}

/* virtual */ objectE
Cclass::type(void) const
{
	return(classE);
}

/* virtual */ void
Cclass::emit_id(void) const
{
	fprintf(stdout, "C%p", this);
}

/* virtual */ void
Cclass::emit_edges(void)
{
	int	i, cnt;

	Centity::emit_edges();

	if (m_extendsP) {
		fprintf(stdout, "E258 C%p C%p\n", this, m_extendsP);
	}
	cnt = m_implements_cnt;
	for (i = 0; i < cnt; ++i) {
		fprintf(stdout, "E259 C%p C%p\n", this, m_implementsPP[i]);
	}
	cnt = m_innerclasses_cnt;
	for (i = 0; i < cnt; ++i) {
		fprintf(stdout, "E260 C%p C%p\n", this, m_innerclassesP[i].m_innerclassP);
	}
	if (m_arrayofP) {
		fprintf(stdout, "E262 C%p C%p\n", this, m_arrayofP);
	}
}

/* virtual */ void
Cclass::emit_instance(void) const
{
	char	c;

	if (m_access_flags & ACC_INTERFACE) {
		c = 'I';
	} else if (m_arrayofP) {
		c = 'A';
	} else {
		c = 'C';
	}

	fprintf(stdout, "$INSTANCE C%p %c\n", this, c);
}

void
Cclass::emit_attributes(void) const
{
	innerclassT	*innerclassP, *endP;
	u2T			access;
	int			dimensions;

	emit_id();
	fputs(" {label=\"", stdout);
	dumpUnicode(m_nameP);
	for (dimensions = m_dimensions; dimensions > 0; --dimensions) {
		fputs("[ ]", stdout);
	}
	fputc('"', stdout);

	fputs(" file=.", stdout);

	if (m_dimensions > 0) {
		fprintf(stdout, " dim=%d\n", m_dimensions);
	}
	if (m_lineno >= 0) {
		fprintf(stdout, " lineno=%d", m_lineno);
	}

	if (m_major_version) {
		fprintf(stdout, " major=%u", (int) m_major_version);
	}
	if (m_minor_version) {
		fprintf(stdout, " minor=%u", (int) m_minor_version);
	}
	if ((access = m_access_flags)) {
		fputs(" access=\"", stdout);
		dump_access(access);
		fputc('"', stdout);
	}
	fputs("}\n", stdout);

	innerclassP = m_innerclassesP;
	for (endP = innerclassP + m_innerclasses_cnt; innerclassP < endP; ++innerclassP) {
		if ((access = innerclassP->m_access)) {
			fprintf(stdout, "(E260 C%p C%p) { access=\"", this, innerclassP->m_innerclassP);
			dump_access(access);
			fputs("\"}\n", stdout);
}	}	}

