#include "typedef.h"

#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include "util.h"
#include "object.h"
#include "edge.h"
#include "variable.h"
#include "field.h"
#include "method.h"
#include "class.h"
#include "directory.h"
#include "source.h"

extern int g_fields;
extern int g_variables;

static void
dump_instances(void)
{
	Csource		*missingP = 0;
	Cdirectory	*directoryP;
	Csource		*sourceP;
	Cclass		*classP;
	Cmethod		*methodP;
	Cfield		*fieldP;
	Cvariable	*variableP;
	unsigned int	i;

	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		directoryP->emit_instance();
	}

	for (i = 0; i < HASHSIZE; ++i) {
		for (sourceP = Csource::g_hash[i]; sourceP; sourceP = sourceP->m_hashP) {
			sourceP->emit_instance();
	}	}

	for (i = 0; i < HASHSIZE; ++i) {
		for (classP = Cclass::g_hash[i]; classP; classP = classP->m_hashP) {
			classP->emit_instance();
	}	}

	for (i = 0; i < HASHSIZE; ++i) {
		for (methodP = Cmethod::g_hash[i]; methodP; methodP = methodP->m_hashP) {
			methodP->emit_instance();
			methodP->add_created();
	}	}

	if (g_fields) {
		for (i = 0; i < HASHSIZE; ++i) {
			for (fieldP = Cfield::g_hash[i]; fieldP; fieldP = fieldP->m_hashP) {
				fieldP->emit_instance();
				fieldP->add_created();
	}	}	}

	if (g_variables) {
		for (i = 0; i < HASHSIZE; ++i) {
			for (variableP = Cvariable::g_hash[i]; variableP; variableP = variableP->m_hashP) {
				variableP->emit_instance();
	}	}	}
}

static void
dump_edges(void)
{
	Cobject		*objectP;

	for (objectP = Cobject::g_created_headP; objectP; objectP = objectP->m_created_nextP) {
		objectP->emit_edges();
	}
}

static void
dump_attributes(void)
{
	Cobject		*objectP;

	fprintf(stdout, "\nFACT ATTRIBUTE :\n\n");

	for (objectP = Cobject::g_created_headP; objectP; objectP = objectP->m_created_nextP) {
		objectP->emit_attributes();
	}
}

static void
dump_schema(void)
{
	extern int		g_hide_edges[NUMBER_EDGE_TYPES];

	fputs(
		"\nSCHEME TUPLE :\n\n"
		"$INHERIT D   $ENTITY\n"	// Directory
		"$INHERIT S   $ENTITY\n"	// Source
		"$INHERIT C   $ENTITY\n"	// Class
		"$INHERIT I   C\n"			// Interface
		"$INHERIT M   $ENTITY\n"	// Method
		"$INHERIT M1  M\n"			// Public method
		"$INHERIT M0  M\n"			// Plain method
		"$INHERIT M4  M\n"			// Protected method
		"$INHERIT M6  M\n"			// Private protected method
		"$INHERIT M2  M\n"			// Private method

		"$INHERIT V   $ENTITY\n"	// Variable
		"$INHERIT A   V\n"			// Array
	, stdout);

	if (g_fields) {
		fputs(
		"$INHERIT F   V\n"			// Field
		"$INHERIT F1  F\n"			// Public field
		"$INHERIT F0  F\n"			// Plain field
		"$INHERIT F4  F\n"			// Protected field
		"$INHERIT F2  F\n"			// Private field
		"$INHERIT FC  F\n"			// Constant field
		,stdout);
	}

	if (g_variables) {
		fputs(
		"$INHERIT L   V\n"			// Local variable
		, stdout);
	}

	fputs(
		"contain  D   D\n" 		// Directory contains Directory
		"contain  D   S\n"		// Directory contains Source
		"contain  S   C\n"		// Source    contains classes
		"contain  C   M\n"		// Class	 contains Methods
	,stdout);

	if (g_fields) {
		fputs(
	    "contain  C   F\n"		// Class     contains Fields
		, stdout);
	}

	if (g_variables) {
		fputs(
		"contain  M   L\n"		// Method    contains Local variables
		,stdout);
	}

	if (!g_hide_edges[EXTENDS_EDGE]) {
		fputs("E258     C   C\n", stdout);		// extends
	}
	if (!g_hide_edges[IMPLEMENTS_INTERFACE_EDGE]) {
		fputs("E259     C   I\n", stdout);		// implements
	}
	if (!g_hide_edges[INNER_EDGE]) {
		fputs("E260     C   C\n", stdout);		// Inner class
	}
	if (!g_hide_edges[THROWS_EDGE]) {
		fputs("E261     M   C\n", stdout);		// Throws
	}
	if (!g_hide_edges[ARRAYOF_EDGE]) {
		fputs("E262     A   C\n"				// Arrayof
			  "E262     A   A\n"
		, stdout);
	}
	if (!g_hide_edges[OVERRIDES_EDGE]) {
		fputs("E256     M   M\n", stdout);		// Overrides
	}
	if (!g_hide_edges[IMPLEMENTS_METHOD_EDGE]) {
		fputs("E257     M   M\n", stdout);		// Interface method
	}

	if (!g_hide_edges[INVOKE_VIRTUAL_EDGE]) {
		fputs("E182     M   M\n", stdout);		// Invoke virtual
	}
	if (!g_hide_edges[INVOKE_SPECIAL_EDGE]) {
		fputs("E183     M   M\n", stdout);		// Invoke special
	}
	if (!g_hide_edges[INVOKE_STATIC_EDGE]) {
		fputs("E184     M   M\n", stdout);		// Invoke static
	}
	if (!g_hide_edges[INVOKE_INTERFACE_EDGE]) {
		fputs("E185     M   I\n", stdout);		// Invoke interface
	}
	if (!g_hide_edges[GET_STATIC_EDGE]) {
		fputs("E178     M   F\n", stdout);		// getstatic
	}
	if (!g_hide_edges[PUT_STATIC_EDGE]) {
		fputs("E179     M   F\n", stdout);		// putstatic
	}
	if (!g_hide_edges[GET_FIELD_EDGE]) {
		fputs("E180     M   F\n", stdout);		// getfield
	}
	if (!g_hide_edges[PUT_FIELD_EDGE]) {
		fputs("E181     M   F\n", stdout);		// putfield
	}
	if (!g_hide_edges[MULTIANEWARRAY_EDGE]) {
		fputs("E197     M   C\n", stdout);		// multianewarray
	}
	if (!g_hide_edges[NEW_EDGE]) {
		fputs("E187     M   C\n", stdout);		// new
	}
	if (!g_hide_edges[ANEWARRAY_EDGE]) {
		fputs("E189     M   C\n", stdout);		// anewarray
	}
	if (!g_hide_edges[CHECKCAST_EDGE]) {
		fputs("E192     M   C\n", stdout);      // checkcast
	}
	if (!g_hide_edges[INSTANCEOF_EDGE]) {
		fputs("E193     M   C\n", stdout);      // instanceof
	}

	if (!g_hide_edges[ILOAD_EDGE]) {
		fputs("E21      M   L\n", stdout);		// iload
	}
	if (!g_hide_edges[LLOAD_EDGE]) {
		fputs("E22      M   L\n", stdout);		// lload
	}
	if (!g_hide_edges[FLOAD_EDGE]) {
		fputs("E23      M   L\n", stdout);		// fload
	}
	if (!g_hide_edges[DLOAD_EDGE]) {
		fputs("E24      M   L\n", stdout);		// dload
	}
	if (!g_hide_edges[ALOAD_EDGE]) {
		fputs("E25      M   L\n", stdout);		// aload
	}

	if (!g_hide_edges[ISTORE_EDGE]) {
		fputs("E54      M   L\n", stdout);		// istore
	}
	if (!g_hide_edges[LSTORE_EDGE]) {
		fputs("E55      M   L\n", stdout);		// lstore
	}
	if (!g_hide_edges[FSTORE_EDGE]) {
		fputs("E56      M   L\n", stdout);		// fstore
	}
	if (!g_hide_edges[DSTORE_EDGE]) {
		fputs("E57      M   L\n", stdout);		// dstore
	}
	if (!g_hide_edges[ASTORE_EDGE]) {
		fputs("E58      M   L\n", stdout);		// astore
	}

	if (!g_hide_edges[IINC_EDGE]) {
		fputs("E132     M   L\n", stdout);		// iinc
	}
	if (!g_hide_edges[JSR_EDGE]) {
		fputs("E168     M   L\n", stdout);		// jsr
	}
	if (!g_hide_edges[RET_EDGE]) {
		fputs("E169     M   L\n", stdout);      // ret
	}
	if (!g_hide_edges[RET_W_EDGE]) {
		fputs("E209     M   L\n", stdout);		// ret_w
	}

	fputs(

		"\nSCHEME ATTRIBUTE :\n\n"
		"$ENTITY {\n"
		"   color       = (255 255 255)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"D {\n"
        "   class_label = \"Directory\"\n"
		"   color       = (255 255 153)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 4\n"
		"}\n"

		"S {\n"
		"   class_label = \"Source\"\n"
		"   color       = (  0 204   0)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"

		"C {\n"
		"   class_label = \"Class\"\n"
		"   color       = (255 204   0)\n"
    	"   labelcolor  = (  0   0   0)\n"
    	"   class_style = 1\n"
	   	"}\n"

		"I {\n"
		"   class_label = \"Interface\"\n"
		"   color       = (239 204 136)\n"
        "   labelcolor  = (  0   0   0)\n"
        "   class_style = 1\n"
		"}\n"

		"M {\n"
		"   class_label = \"Method\"\n"
		"   class_style = 1\n"
		"}\n"

		"M1 {\n"
		"   class_label = \"Public method\"\n"
		"   color       = (255 102 102)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"M0 {\n"
		"   class_label = \"Plain method\"\n"
		"   color       = (218 102 102)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"M4 {\n"
		"   class_label = \"Protected method\"\n"
		"   color       = (255 153 153)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"M6 {\n"
		"   class_label = \"Private Protected method\"\n"
		"   color       = (221   0 204)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"M2 {\n"
		"   class_label = \"Private method\"\n"
		"   color       = (255 156 255)\n"
		"   labelcolor  = (  0   0   0)\n"
		"   class_style = 1\n"
		"}\n"

		"V {\n"
		"   class_label = \"Variable\"\n"
  		"   class_style = 6\n"
  		"}\n"

		"A {\n"
		"   class_label = \"Array\"\n"
		"   color       = (204 255 255)\n"
		"   class_style = 5\n"
  		"}\n"
	,stdout);

	if (g_fields) {
		fputs(
		"F {\n"
		"  class_label  = \"Field\"\n"
  		"  class_style = 6\n"
  		"}\n"

		"F1 {\n"
		"  class_label  = \"Public Field\"\n"
  		"  color       = (  0   0 255)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"

		"F0 {\n"
		"  class_label  = \"Plain Field\"\n"
  		"  color       = (  0 102 153)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"

		"F4 {\n"
		"  class_label  = \"Protected Field\"\n"
  		"  color       = (111 101 255)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"

		"F2 {\n"
		"  class_label  = \"Private Field\"\n"
  		"  color       = ( 35 174 255)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"

		"FC {\n"
		"  class_label  = \"Constant\"\n"
  		"  color       = (  0 204 198)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"
		, stdout);
	}

	if (g_variables) {
		fputs(

		"L {\n"
		"  class_label  = \"Local variable\"\n"
  		"  color       = (204 204 255)\n"
  		"  labelcolor  = (  0   0   0)\n"
  		"  class_style = 6\n"
  		"}\n"
		, stdout);
	}

	if (!g_hide_edges[EXTENDS_EDGE]) {
		fputs(
		"(E258) {\n"
		"   class_label = \"Extends class\"\n"
		"   color       = (255 255   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[IMPLEMENTS_INTERFACE_EDGE]) {
		fputs(
		"(E259) {\n"
		"   class_label = \"Implements interface\"\n"
		"   color       = (204 204   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INNER_EDGE]) {
		fputs(
		"(E260) {\n"
		"   class_label = \"Inner class\"\n"
		"   color       = (102 153   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[THROWS_EDGE]) {
		fputs(
		"(E261) {\n"
		"   class_label = \"Throws class\"\n"
		"   color       = (153 153   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ARRAYOF_EDGE]) {
		fputs(
		"(E262) {\n"
		"   class_label = \"Array of\"\n"
		"   color       = (153 153   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[OVERRIDES_EDGE]) {
		fputs(
		"(E256)   {\n"
		"   class_label = \"Overrides method\"\n"
		"   color       = (153   0 204)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[IMPLEMENTS_METHOD_EDGE]) {
		fputs(
		"(E257)   {\n"
		"   class_label = \"Interface method\"\n"
		"   color       = (153   0 204)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INVOKE_VIRTUAL_EDGE]) {
		fputs(
		"(E182) {\n"
		"   class_label = \"Invoke virtual\"\n"
		"   color       = (153   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INVOKE_SPECIAL_EDGE]) {
		fputs(
		"(E183) {\n"
		"   class_label = \"Invoke special\"\n"
		"   color       = (153   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INVOKE_STATIC_EDGE]) {
		fputs(
		"(E184) {\n"
		"   class_label = \"Invoke static\"\n"
		"   color       = (153   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INVOKE_INTERFACE_EDGE]) {
		fputs(
		"(E185) {\n"
		"   class_label = \"Invoke interface\"\n"
		"   color       = (153   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[GET_STATIC_EDGE]) {
		fputs(
		"(E178) {\n"
		"   class_label = \"get static\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[PUT_STATIC_EDGE]) {
		fputs(
		"(E179) {\n"
		"   class_label = \"put static\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[GET_FIELD_EDGE]) {
		fputs(
		"(E180) {\n"
		"   class_label = \"get field\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[PUT_FIELD_EDGE]) {
		fputs(
		"(E181) {\n"
		"   class_label = \"put field\"\n"
		"   color       = (  0  0  0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[MULTIANEWARRAY_EDGE]) {
		fputs(
		"(E197) {\n"
		"   class_label = \"new multi array\"\n"
		"   color       = (  0 102   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[NEW_EDGE]) {
		fputs(
		"(E187) {\n"
		"   class_label = \"new\"\n"
		"   color       = (  0 102   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ANEWARRAY_EDGE]) {
		fputs(
		"(E189) {\n"
		"   class_label = \"new array\"\n"
		"   color       = (  0 102   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[CHECKCAST_EDGE]) {
		fputs(
		"(E192) {\n"
		"   class_label = \"check cast\"\n"
		"   color       = (  0 159   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[INSTANCEOF_EDGE]) {
		fputs(
		"(E193) {\n"
		"   class_label = \"instanceof\"\n"
		"   color       = (  0 159   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ILOAD_EDGE]) {
		fputs(
		"(E21) {\n"
		"   class_label = \"load int\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[LLOAD_EDGE]) {
		fputs(
		"(E22) {\n"
		"   class_label = \"load long\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[FLOAD_EDGE]) {
		fputs(
		"(E23) {\n"
		"   class_label = \"load float\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[DLOAD_EDGE]) {
		fputs(
		"(E24) {\n"
		"   class_label = \"load double\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ALOAD_EDGE]) {
		fputs(
		"(E25) {\n"
		"   class_label = \"load ref\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ISTORE_EDGE]) {
		fputs(
		"(E54) {\n"
		"   class_label = \"store int\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[LSTORE_EDGE]) {
		fputs(
		"(E55) {\n"
		"   class_label = \"store long\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[FSTORE_EDGE]) {
		fputs(
		"(E56) {\n"
		"   class_label = \"store float\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[DSTORE_EDGE]) {
		fputs(
		"(E57) {\n"
		"   class_label = \"store double\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[ASTORE_EDGE]) {
		fputs(
		"(E58) {\n"
		"   class_label = \"store ref\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[IINC_EDGE]) {
		fputs(
		"(E132) {\n"
		"   class_label = \"increment\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[JSR_EDGE]) {
		fputs(
		"(E168) {\n"
		"   class_label = \"call subroutine\"\n"
		"   color       = (  0   0   0)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[RET_EDGE]) {
		fputs(
		"(E169) {\n"
		"   class_label = \"return\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	if (!g_hide_edges[RET_W_EDGE]) {
		fputs(
		"(E209) {\n"
		"   class_label = \"return wide\"\n"
		"   color       = (167 142  38)\n"
		"   class_style = 0\n"
		"}\n"
		, stdout);
	}

	fputs(
		"\nFACT TUPLE :\n\n"
	, stdout);
}

static void
collapse_directories(void)
{
	Cdirectory	*directoryP, *parentP;
	Cdirectory	*collapseP;

	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		if ((parentP = (Cdirectory *) directoryP->m_parentP)) {
			parentP->m_collapse++;
	}	}
	
	collapseP = 0;
	for (directoryP = Cdirectory::g_headP; directoryP; directoryP = directoryP->m_nextP) {
		parentP = (Cdirectory *) directoryP->m_parentP;
		if (!parentP) {
			// No point collapsing a directory that is already the root
			continue;
		}
		for (;; parentP = (Cdirectory *) parentP->m_parentP) {
			if (!parentP) {
				if (!collapseP) {
					collapseP = directoryP;
//					fprintf(stderr, "Initial collapse %s\n", collapseP->m_nameP);
				}
				break;
			}
			if (parentP->m_collapse > 1) {
				break;
			}
			if (parentP == collapseP) {
				collapseP = directoryP;
//				fprintf(stderr, "New collapse %s\n", collapseP->m_nameP);
	}	}	}
		
	if (collapseP) {

		char		buffer[2048];
		size_t		lth;

		lth = collapseP->full_path(buffer);
		buffer[--lth] = 0;
		collapseP->name(buffer);

		for (parentP = collapseP; parentP = (Cdirectory *) parentP->m_parentP; ) {
			parentP->m_collapse = -2;
		}
		collapseP->m_parentP = 0;
}	}

static void
find_overrides(void)
{
	Cmethod			*methodP, *method1P, *bestP;
	const u2T		*nameP, *descriptorP;
	const Cclass	*classP, *class1P;
	const Cclass	*extendsP;
	u4T				best_depth, depth, i;

	for (i = 0; i < HASHSIZE; ++i) {
		for (methodP = Cmethod::g_hash[i]; methodP; methodP = methodP->m_hashP) {
			nameP       = methodP->m_nameP;
			descriptorP = methodP->m_descriptorP;
			classP      = (const Cclass *) methodP->m_parentP;
			bestP       = 0;
			best_depth  = 0xFFFFFFFF;
			for (method1P = methodP->g_hash[i]; method1P; method1P = method1P->m_hashP) {
				if (method1P == methodP) {
					continue;
				}
				if (unicodecmp(method1P->m_nameP, nameP)) {
					continue;
				}
				if (unicodecmp(method1P->m_descriptorP, descriptorP)) {
					continue;
				}
				class1P = (const Cclass *) method1P->m_parentP;

				depth = 0;
				for (extendsP = classP->m_extendsP; extendsP; extendsP = extendsP->m_extendsP) {
					if (extendsP == class1P) {
						if (depth < best_depth) {
							bestP      = method1P;
							best_depth = depth;
						}
						break;
					}
					++depth;
				}
				if (best_depth == 0) {
					break;
			}	}
			if (bestP) {
				Cedge::locate(methodP, bestP, OVERRIDES_EDGE, methodP->m_lineno);
}	}	}	}

static void
find_interface_methods(void)
{
	Cmethod			*methodP, *method1P;
	const u2T		*nameP, *descriptorP;
	const Cclass	*classP, *class1P;
	int				i, j;

	for (i = 0; i < HASHSIZE; ++i) {
		for (methodP = Cmethod::g_hash[i]; methodP; methodP = methodP->m_hashP) {
			nameP       = methodP->m_nameP;
			descriptorP = methodP->m_descriptorP;
			classP      = (const Cclass *) methodP->m_parentP;
			for (method1P = methodP->g_hash[i]; method1P; method1P = method1P->m_hashP) {
				if (method1P == methodP) {
					continue;
				}
				if (unicodecmp(method1P->m_nameP, nameP)) {
					continue;
				}
				if (unicodecmp(method1P->m_descriptorP, descriptorP)) {
					continue;
				}
				class1P = (const Cclass *) method1P->m_parentP;

				if (!(class1P->m_access_flags & ACC_INTERFACE)) {
					continue;
				}

				for (j = classP->m_implements_cnt; --j >= 0; ) {
					if (classP->m_implementsPP[j] == class1P) {
						Cedge::locate(methodP, method1P, IMPLEMENTS_METHOD_EDGE, methodP->m_lineno);
						break;
}	}	}	}	}	}

static void
group_datatypes(void)
{
	Csource		*datatypesP = 0;
	int			i;
	Cclass		*classP;

	for (i = 0; i < HASHSIZE; ++i) {
		for (classP = Cclass::g_hash[i]; classP; classP = classP->m_hashP) {
			if (classP->m_parentP == ((Centity *) -1)) {
				if (!datatypesP) {
					datatypesP = Csource::locate(":datatypes:");
				}
				classP->set_parent(datatypesP);
}	}	}	}

static void
deslash_classnames(void)
{
	Cclass		*classP;
	int			i;

	for (i = 0; i < HASHSIZE; ++i) {
		for (classP = Cclass::g_hash[i]; classP; classP = classP->m_hashP) {
			deslash(classP->m_nameP);
}	}	}


void
emitta(void)
{
	group_datatypes();
	find_overrides();
	find_interface_methods();
	deslash_classnames();
	collapse_directories();

	dump_schema();
	dump_instances();
	dump_edges();
	dump_attributes();
}

