#include "typedef.h"

#include <stdio.h>
#include <string.h>

#include "util.h"
#include "object.h"
#include "field.h"

Cfield	*Cfield::g_hash[HASHSIZE] = {0};

/* static */ u4T
Cfield::hash(Centity *entityP, const u2T *nameP)
{
	int		ret  = (int) entityP;
	u2T		c;
	const	u2T *P;

	for (P = nameP; c = *P ; ++P) {
		ret  += c;
		ret <<= 1;
		if (ret & HASHSIZE) {
			ret |= 1;
	}	}
	return(ret & (HASHSIZE-1));
}

Cfield::Cfield(Centity *parentP, const u2T *nameP, const u2T *descriptorP)
{
	m_hashP          = 0;
	m_parentP        = parentP;
	m_nameP          = unicodedup(nameP);
	m_descriptorP    = unicodedup(descriptorP);
	m_access_flags   = 0;
	m_constant.tag   = CONSTANT_Null;
}

Cfield *
Cfield::locate(Centity *parentP, const u2T *nameP, const u2T *descriptorP)
{
	unsigned int	index;
	Cfield			**fieldPP, *fieldP;

	index         = hash(parentP, nameP);
	for (fieldPP = g_hash + index; fieldP = *fieldPP; fieldPP = &(fieldP->m_hashP)) {
		if (fieldP->m_parentP != parentP) {
			continue;
		}
		if (unicodecmp(fieldP->m_nameP, nameP)) {
			continue;
		}
		return(fieldP);
	}	
	fieldP = new Cfield(parentP, nameP, descriptorP);
	if (!fieldP) {
		outofmemory();
	}
	*fieldPP = fieldP;
	return(fieldP);
}

/* virtual */ objectE
Cfield::type(void) const
{
	return(fieldE);
}

/* virtual */ void
Cfield::emit_id(void) const
{
	fprintf(stdout, "F%p", this);
}

const char *
Cfield::get_type(void) const
{
	u2T			access = m_access_flags;
	const char 	*typeP;

	if (access & ACC_FINAL) {
		typeP = "FC";
	} else if (access & ACC_PRIVATE) {
		typeP = "F2";
	} else if (access & ACC_PROTECTED) {
		typeP = "F4";
	} else if (access & ACC_PUBLIC) {
		typeP = "F1";
	} else {
		typeP = "F0";
	}
	return (typeP);
}
		
/* virtual */ void
Cfield::emit_instance(void) const
{
	fprintf(stdout, "$INSTANCE F%p %s\n", this, get_type());
}

void
Cfield::emit_attributes(void) const
{
	const cp_infoT	*constantP;

	emit_id();
	fputs(" {label=\"", stdout);
	dumpUnicode(m_nameP);
	fputc('"', stdout);

	fputs(" file=.", stdout);

	if (m_descriptorP) {
		fputs(" descriptor=\"", stdout);
		dump_descriptor(m_descriptorP);
		fputc('"', stdout);
	}

	if (m_access_flags) {
		fputs(" access=\"", stdout);
		dump_access(m_access_flags);
		fputc('"', stdout);
	}
	
	constantP = &m_constant;

	switch (constantP->tag) {
	case CONSTANT_String:
		fputs(" value=\"", stdout);
		dumpUnicode(constantP->u.stringP);
		fputc('"', stdout);
		break;
	case CONSTANT_Integer:
		printf(" value=%d", constantP->u.ival);
		break;
	case CONSTANT_Float:
		printf(" value=%f", constantP->u.fval);
		break;
	case CONSTANT_Long:
		printf(" value=%I64d", constantP->u.lval);
		break;
	case CONSTANT_Double:
		printf(" value=%lf", constantP->u.dval);
		break;
	}
	fputs("}\n", stdout);
}
